from enum import Enum


class HTTPMethod(str, Enum):
    """
    HTTP Methods
    """

    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"


class TxnType(str, Enum):
    """
    Transaction type
    """

    # Request from API, usually e-commerce
    E_COMMERCE = "e_commerce"

    # Merchant requests for a payment to somebody
    PAYMENT_REQUEST = "payment_request"



import json
import logging

from django.http import JsonResponse
from django.views.generic.base import ContextMixin, View

from ...errors import WebhookProcessingError
from ...utils import verify_signature
from . import conf
from .models import Webhook

logger = logging.getLogger("ottu-py")


class WebhookViewAbstractView(ContextMixin, View):
    WebHookError = WebhookProcessingError
    WebHookModel = Webhook
    status_codes = {
        "success": 200,
        "failure": 400,
        "unverified": 401,
    }
    _data = None

    @property
    def data(self):
        if self._data is None:
            self._data = self.request.POST or json.loads(
                self.request.body.decode("utf-8"),
            )
        return self._data

    def verify(self) -> bool:
        signature_server = self.data.get("signature", "")
        return verify_signature(
            payload=self.data,
            signature=signature_server,
            webhook_key=conf.WEBHOOK_KEY,
        )

    def clean_data(self, processed_data):
        # Use either `processed_data` or `self.data` to create a dict
        return self.data

    def process_data(self):
        # Do something with `self.data`
        return self.data

    def save_data(self, processed_data):
        cleaned_data = self.clean_data(processed_data)
        instance = self.WebHookModel.create_from_webhook(data=cleaned_data)
        return instance

    def post(self, request, *args, **kwargs):
        logger.info(f"Webhook received: {self.data}")
        verified = self.verify()
        if not verified:
            return JsonResponse(
                data={"detail": "Unable to verify signature"},
                status=self.status_codes["unverified"],
            )
        try:
            processed_data = self.process_data()
            self.save_data(processed_data=processed_data)
            return JsonResponse(
                data={"detail": "Success"},
                status=self.status_codes["success"],
            )
        except self.WebHookError:
            return JsonResponse(
                data={"detail": "Failed to process webhook"},
                status=self.status_codes["failure"],
            )



import logging
import typing
from dataclasses import dataclass

from .decorators import interruption_handler
from .enums import HTTPMethod, TxnType
from .errors import APIInterruptError, ValidationError
from .mixins import AsDictMixin
from .request import OttuPYResponse
from .utils import remove_empty_values

if typing.TYPE_CHECKING:
    from .ottu import Ottu

logger = logging.getLogger("ottu-py")


@dataclass
class PaymentMethod(AsDictMixin):
    code: typing.Optional[str] = None
    name: typing.Optional[str] = None
    pg: typing.Optional[str] = None
    type: typing.Optional[str] = None
    amount: typing.Optional[str] = None
    currency_code: typing.Optional[str] = None
    fee: typing.Optional[str] = None
    fee_description: typing.Optional[str] = None
    icon: typing.Optional[str] = None
    flow: typing.Optional[str] = None
    redirect_url: typing.Optional[str] = None

    def __repr__(self):
        return f"PaymentMethod({self.code or '######'})"


class Session:
    url_session_create = "/b/checkout/v1/pymt-txn/"
    url_ops = "/b/pbl/v2/operation/"
    url_auto_debit = "/b/pbl/v2/auto-debit/"

    amount: typing.Optional[str] = None
    attachment: typing.Optional[str] = None
    attachment_short_url: typing.Optional[str] = None
    billing_address: typing.Optional[dict] = None
    checkout_short_url: typing.Optional[str] = None
    checkout_url: typing.Optional[str] = None
    currency_code: typing.Optional[str] = None
    customer_email: typing.Optional[str] = None
    customer_first_name: typing.Optional[str] = None
    customer_last_name: typing.Optional[str] = None
    customer_id: typing.Optional[str] = None
    customer_phone: typing.Optional[str] = None
    due_datetime: typing.Optional[str] = None
    email_recipients: typing.Optional[list[str]] = None
    expiration_time: typing.Optional[str] = None
    extra: typing.Optional[dict] = None
    initiator_id: typing.Optional[int] = None
    language: typing.Optional[str] = None
    mode: typing.Optional[str] = None
    notifications: typing.Optional[dict] = None
    operation: typing.Optional[str] = None
    order_no: typing.Optional[str] = None
    payment_methods: typing.Optional[list[PaymentMethod]] = None
    pg_codes: typing.Optional[list[str]] = None
    qr_code_url: typing.Optional[str] = None
    redirect_url: typing.Optional[str] = None
    session_id: typing.Optional[str] = None
    shipping_address: typing.Optional[dict] = None
    state: typing.Optional[str] = None
    type: typing.Optional[str] = None
    vendor_name: typing.Optional[str] = None
    webhook_url: typing.Optional[str] = None

    def __init__(self, ottu: "Ottu", **data):
        self.ottu = ottu
        for field, value in data.items():
            setattr(self, field, value)

        payment_methods = getattr(self, "payment_methods", [])
        if payment_methods:
            self.payment_methods = [
                PaymentMethod(
                    **payment_method,
                )
                for payment_method in payment_methods
            ]

    def __repr__(self):
        return f"Session({self.session_id or '######'})"

    def __bool__(self):
        return bool(self.session_id)

    def as_dict(self):
        fields = [
            "amount",
            "attachment",
            "attachment_short_url",
            "billing_address",
            "checkout_short_url",
            "checkout_url",
            "currency_code",
            "customer_email",
            "customer_first_name",
            "customer_last_name",
            "customer_id",
            "customer_phone",
            "due_datetime",
            "email_recipients",
            "expiration_time",
            "extra",
            "initiator_id",
            "language",
            "mode",
            "notifications",
            "operation",
            "order_no",
            "payment_methods",
            "pg_codes",
            "qr_code_url",
            "redirect_url",
            "session_id",
            "shipping_address",
            "state",
            "type",
            "vendor_name",
            "webhook_url",
        ]
        return remove_empty_values(
            {field: getattr(self, field, "") for field in fields},
        )

    def __path_to_file(self, path: str) -> typing.Tuple[str, bytes]:
        with open(path, "rb") as f:
            content = f.read()
            name = f.name or "attachment.pdf"
        return name, content

    def create(
        self,
        *,
        txn_type: TxnType,
        amount: str,
        currency_code: str,
        pg_codes: list[str],
        payment_type: str = "one_off",
        customer_id: typing.Optional[str] = None,
        customer_email: typing.Optional[str] = None,
        customer_phone: typing.Optional[str] = None,
        customer_first_name: typing.Optional[str] = None,
        customer_last_name: typing.Optional[str] = None,
        agreement: typing.Optional[dict] = None,
        card_acceptance_criteria: typing.Optional[dict] = None,
        attachment: typing.Optional[str] = None,
        billing_address: typing.Optional[dict] = None,
        due_datetime: typing.Optional[str] = None,
        email_recipients: typing.Optional[list[str]] = None,
        expiration_time: typing.Optional[str] = None,
        extra: typing.Optional[dict] = None,
        generate_qr_code: typing.Optional[bool] = None,
        language: typing.Optional[str] = None,
        mode: typing.Optional[str] = None,
        notifications: typing.Optional[dict] = None,
        order_no: typing.Optional[str] = None,
        product_type: typing.Optional[str] = None,
        redirect_url: typing.Optional[str] = None,
        shopping_address: typing.Optional[dict] = None,
        shortify_attachment_url: typing.Optional[bool] = None,
        shortify_checkout_url: typing.Optional[bool] = None,
        vendor_name: typing.Optional[str] = None,
        webhook_url: typing.Optional[str] = None,
        include_sdk_setup_preload: typing.Optional[bool] = None,
        **kwargs,
    ) -> dict:
        """
        Creates a new checkout session.
        :param txn_type: Transaction type
        :param amount: Amount
        :param currency_code: Currency code
        :param pg_codes: Payment gateway codes
        :param customer_id: Customer ID
        :param customer_email: Customer email
        :param customer_phone: Customer phone
        :param customer_first_name: Customer first name
        :param customer_last_name: Customer last name
        :param agreement: Agreement
        :param card_acceptance_criteria: Card acceptance criteria
        :param attachment: Path to attachment
        :param billing_address: Billing address
        :param due_datetime: Due datetime
        :param email_recipients: Email recipients
        :param expiration_time: Expiration time
        :param extra: Extra
        :param generate_qr_code: Generate QR code
        :param language: Language
        :param mode: Mode
        :param notifications: Notifications
        :param order_no: Order number
        :param product_type: Product type
        :param redirect_url: Redirect URL
        :param shopping_address: Shopping address
        :param shortify_attachment_url: Shortify attachment URL
        :param shortify_checkout_url: Shortify checkout URL
        :param vendor_name: Vendor name
        :param webhook_url: Webhook URL
        :param include_sdk_setup_preload: Include SDK setup preload
        :param kwargs: Additional arguments supported by the API
        :return: Session
        """
        if kwargs:
            msg = (
                f"The following arguments are not "
                f"supported by the SDK: {', '.join(kwargs.keys())}"
            )
            logger.warning(msg)

        customer_id = customer_id or self.ottu.customer_id
        payload = {
            "type": txn_type.value,
            "amount": amount,
            "currency_code": currency_code,
            "pg_codes": pg_codes,
            "payment_type": payment_type,
            "customer_id": customer_id,
            "customer_email": customer_email,
            "customer_phone": customer_phone,
            "customer_first_name": customer_first_name,
            "customer_last_name": customer_last_name,
            "agreement": agreement,
            "card_acceptance_criteria": card_acceptance_criteria,
            "billing_address": billing_address,
            "due_datetime": due_datetime,
            "email_recipients": email_recipients,
            "expiration_time": expiration_time,
            "extra": extra,
            "generate_qr_code": generate_qr_code,
            "language": language,
            "mode": mode,
            "notifications": notifications,
            "order_no": order_no,
            "product_type": product_type,
            "redirect_url": redirect_url,
            "shopping_address": shopping_address,
            "shortify_attachment_url": shortify_attachment_url,
            "shortify_checkout_url": shortify_checkout_url,
            "vendor_name": vendor_name,
            "webhook_url": webhook_url,
            "include_sdk_setup_preload": include_sdk_setup_preload,
        }
        payload = remove_empty_values(payload)
        payload.update(kwargs)  # `kwargs` may contain `None` values
        if attachment:
            json_or_form = {
                "data": payload,
                "files": {"attachment": self.__path_to_file(path=attachment)},
            }
        else:
            json_or_form = {
                "json": payload,
            }
        ottu_py_response = self.ottu.send_request(
            path=self.url_session_create,
            method=HTTPMethod.POST,
            **json_or_form,
        )
        session = Session(
            ottu=self.ottu,
            **ottu_py_response.response,
        )
        if ottu_py_response.success:
            self.ottu._update_session(session)
        return ottu_py_response.as_dict()

    def retrieve(self, session_id: str) -> dict:
        """
        Retrieves a checkout session.
        :param session_id: Session ID
        """
        ottu_py_response = self.ottu.send_request(
            path=f"{self.url_session_create}{session_id}",
            method=HTTPMethod.GET,
        )
        session = Session(
            ottu=self.ottu,
            **ottu_py_response.response,
        )
        if ottu_py_response.success:
            self.ottu._update_session(session)
        return ottu_py_response.as_dict()

    def refresh(self, session_id: typing.Optional[str] = None) -> typing.Optional[dict]:
        """
        Reloads the payment attributes from upstream by calling the `retrieve` method.
        """
        session_id = session_id or self.session_id
        if session_id:
            response = self.retrieve(session_id=session_id)
            if response["success"]:
                return response
        return None

    def update(
        self,
        *,
        amount: typing.Optional[str] = None,
        currency_code: typing.Optional[str] = None,
        pg_codes: typing.Optional[list[str]] = None,
        customer_id: typing.Optional[str] = None,
        customer_email: typing.Optional[str] = None,
        customer_phone: typing.Optional[str] = None,
        customer_first_name: typing.Optional[str] = None,
        customer_last_name: typing.Optional[str] = None,
        attachment: typing.Optional[str] = None,
        billing_address: typing.Optional[dict] = None,
        due_datetime: typing.Optional[str] = None,
        email_recipients: typing.Optional[list[str]] = None,
        expiration_time: typing.Optional[str] = None,
        extra: typing.Optional[dict] = None,
        generate_qr_code: typing.Optional[bool] = None,
        language: typing.Optional[str] = None,
        mode: typing.Optional[str] = None,
        notifications: typing.Optional[dict] = None,
        order_no: typing.Optional[str] = None,
        product_type: typing.Optional[str] = None,
        redirect_url: typing.Optional[str] = None,
        shopping_address: typing.Optional[dict] = None,
        shortify_attachment_url: typing.Optional[bool] = None,
        shortify_checkout_url: typing.Optional[bool] = None,
        vendor_name: typing.Optional[str] = None,
        webhook_url: typing.Optional[str] = None,
    ) -> dict:
        payload = {
            "amount": amount,
            "currency_code": currency_code,
            "pg_codes": pg_codes,
            "customer_id": customer_id,
            "customer_email": customer_email,
            "customer_phone": customer_phone,
            "customer_first_name": customer_first_name,
            "customer_last_name": customer_last_name,
            "billing_address": billing_address,
            "due_datetime": due_datetime,
            "email_recipients": email_recipients,
            "expiration_time": expiration_time,
            "extra": extra,
            "generate_qr_code": generate_qr_code,
            "language": language,
            "mode": mode,
            "notifications": notifications,
            "order_no": order_no,
            "product_type": product_type,
            "redirect_url": redirect_url,
            "shopping_address": shopping_address,
            "shortify_attachment_url": shortify_attachment_url,
            "shortify_checkout_url": shortify_checkout_url,
            "vendor_name": vendor_name,
            "webhook_url": webhook_url,
        }
        payload = remove_empty_values(payload)
        if attachment:
            json_or_form = {
                "data": payload,
                "files": {"attachment": self.__path_to_file(path=attachment)},
            }
        else:
            json_or_form = {
                "json": payload,
            }
        ottu_py_response = self.ottu.send_request(
            path=f"{self.url_session_create}{self.session_id}",
            method=HTTPMethod.PATCH,
            **json_or_form,
        )
        session = Session(ottu=self.ottu, **ottu_py_response.response)
        if ottu_py_response.success:
            self.ottu._update_session(session)
        return ottu_py_response.as_dict()

    def auto_debit(self, token: str, session_id: str) -> dict:
        payload = {
            "session_id": session_id,
            "token": token,
        }
        ottu_py_response = self.ottu.send_request(
            path=self.url_auto_debit,
            method=HTTPMethod.POST,
            json=payload,
        )
        return ottu_py_response.as_dict()

    def ops(
        self,
        operation: str,
        order_id: typing.Optional[str] = None,
        session_id: typing.Optional[str] = None,
        amount: typing.Optional[str] = None,
        headers: typing.Optional[dict] = None,
    ) -> OttuPYResponse:
        if session_id is None:
            session_id = self.session_id

        if not session_id and not order_id:
            raise ValidationError("session_id or order_id is required")

        payload = {
            "session_id": session_id,
            "order_no": order_id,
            "operation": operation,
            "amount": amount,
        }
        payload = remove_empty_values(payload)
        return self.ottu.send_request(
            path=self.url_ops,
            method=HTTPMethod.POST,
            json=payload,
            headers=headers,
        )

    def cancel(
        self,
        order_id: typing.Optional[str] = None,
        session_id: typing.Optional[str] = None,
    ) -> dict:
        ottu_py_response = self.ops(
            operation="cancel",
            order_id=order_id,
            session_id=session_id,
        )
        if ottu_py_response.success:
            self.refresh()
        return ottu_py_response.as_dict()

    def expire(
        self,
        order_id: typing.Optional[str] = None,
        session_id: typing.Optional[str] = None,
    ) -> dict:
        ottu_py_response = self.ops(
            operation="expire",
            order_id=order_id,
            session_id=session_id,
        )
        if ottu_py_response.success:
            self.refresh()
        return ottu_py_response.as_dict()

    def delete(
        self,
        order_id: typing.Optional[str] = None,
        session_id: typing.Optional[str] = None,
    ) -> dict:
        ottu_py_response = self.ops(
            operation="delete",
            order_id=order_id,
            session_id=session_id,
        )
        return ottu_py_response.as_dict()

    def capture(
        self,
        order_id: typing.Optional[str] = None,
        session_id: typing.Optional[str] = None,
        amount: typing.Optional[str] = None,
        tracking_key: typing.Optional[str] = None,
    ) -> dict:
        headers = None
        if tracking_key:
            headers = {
                "Tracking-Key": tracking_key,
            }
        ottu_py_response = self.ops(
            operation="capture",
            order_id=order_id,
            session_id=session_id,
            amount=amount,
            headers=headers,
        )
        if ottu_py_response.success:
            self.refresh()
        return ottu_py_response.as_dict()

    def refund(
        self,
        order_id: typing.Optional[str] = None,
        session_id: typing.Optional[str] = None,
        amount: typing.Optional[str] = None,
        tracking_key: typing.Optional[str] = None,
    ) -> dict:
        headers = None
        if tracking_key:
            headers = {
                "Tracking-Key": tracking_key,
            }
        ottu_py_response = self.ops(
            operation="refund",
            order_id=order_id,
            session_id=session_id,
            amount=amount,
            headers=headers,
        )
        if ottu_py_response.success:
            self.refresh()
        return ottu_py_response.as_dict()

    def void(
        self,
        order_id: typing.Optional[str] = None,
        session_id: typing.Optional[str] = None,
        tracking_key: typing.Optional[str] = None,
    ) -> dict:
        headers = None
        if tracking_key:
            headers = {
                "Tracking-Key": tracking_key,
            }
        ottu_py_response = self.ops(
            operation="void",
            order_id=order_id,
            session_id=session_id,
            headers=headers,
        )
        if ottu_py_response.success:
            self.refresh()
        return ottu_py_response.as_dict()

    def get_pg_codes(self, plugin, currency, tokenizable=False) -> list:
        if self.payment_methods:
            return [pm.code for pm in self.payment_methods]

        response = self.ottu.get_payment_methods(
            plugin=plugin,
            currencies=[
                currency,
            ],
            tokenizable=tokenizable,
        )
        if not response["success"]:
            raise APIInterruptError(**response)
        return [pm["code"] for pm in response["response"]["payment_methods"]]

    def get_auto_debit_pg_codes(self, plugin, currency) -> list:
        # There is no way to identify the
        # cached payment method supports auto debit or not.
        # So, we are calling the API again.
        response = self.ottu.get_payment_methods(
            plugin=plugin,
            currencies=[
                currency,
            ],
            tokenizable=True,
        )
        if not response["success"]:
            raise APIInterruptError(**response)
        return [pm["code"] for pm in response["response"]["payment_methods"]]

    def get_token_from_db(self, agreement, customer_id) -> str:
        raise NotImplementedError("Please implement this method in your subclass")

    @interruption_handler
    def checkout_autoflow(
        self,
        *,
        txn_type: TxnType,
        amount: str,
        currency_code: str,
        payment_type: str = "one_off",
        customer_id: typing.Optional[str] = None,
        customer_email: typing.Optional[str] = None,
        customer_phone: typing.Optional[str] = None,
        customer_first_name: typing.Optional[str] = None,
        customer_last_name: typing.Optional[str] = None,
        agreement: typing.Optional[dict] = None,
        card_acceptance_criteria: typing.Optional[dict] = None,
        attachment: typing.Optional[str] = None,
        billing_address: typing.Optional[dict] = None,
        due_datetime: typing.Optional[str] = None,
        email_recipients: typing.Optional[list[str]] = None,
        expiration_time: typing.Optional[str] = None,
        extra: typing.Optional[dict] = None,
        generate_qr_code: typing.Optional[bool] = None,
        language: typing.Optional[str] = None,
        mode: typing.Optional[str] = None,
        notifications: typing.Optional[dict] = None,
        order_no: typing.Optional[str] = None,
        product_type: typing.Optional[str] = None,
        redirect_url: typing.Optional[str] = None,
        shopping_address: typing.Optional[dict] = None,
        shortify_attachment_url: typing.Optional[bool] = None,
        shortify_checkout_url: typing.Optional[bool] = None,
        vendor_name: typing.Optional[str] = None,
        webhook_url: typing.Optional[str] = None,
        include_sdk_setup_preload: typing.Optional[bool] = None,
        checkout_extra_args: typing.Optional[dict] = None,
    ):
        pg_codes = self.get_pg_codes(plugin=txn_type, currency=currency_code)
        checkout_extra_args = checkout_extra_args or {}
        return self.create(
            txn_type=txn_type,
            amount=amount,
            currency_code=currency_code,
            pg_codes=pg_codes,
            payment_type=payment_type,
            customer_id=customer_id,
            customer_email=customer_email,
            customer_phone=customer_phone,
            customer_first_name=customer_first_name,
            customer_last_name=customer_last_name,
            agreement=agreement,
            card_acceptance_criteria=card_acceptance_criteria,
            attachment=attachment,
            billing_address=billing_address,
            due_datetime=due_datetime,
            email_recipients=email_recipients,
            expiration_time=expiration_time,
            extra=extra,
            generate_qr_code=generate_qr_code,
            language=language,
            mode=mode,
            notifications=notifications,
            order_no=order_no,
            product_type=product_type,
            redirect_url=redirect_url,
            shopping_address=shopping_address,
            shortify_attachment_url=shortify_attachment_url,
            shortify_checkout_url=shortify_checkout_url,
            vendor_name=vendor_name,
            webhook_url=webhook_url,
            include_sdk_setup_preload=include_sdk_setup_preload,
            **checkout_extra_args,
        )

    @interruption_handler
    def auto_debit_autoflow(
        self,
        *,
        txn_type: TxnType,
        amount: str,
        currency_code: str,
        customer_id: str,
        agreement: dict,
        pg_codes: typing.Optional[list[str]] = None,
        customer_email: typing.Optional[str] = None,
        customer_phone: typing.Optional[str] = None,
        customer_first_name: typing.Optional[str] = None,
        customer_last_name: typing.Optional[str] = None,
        card_acceptance_criteria: typing.Optional[dict] = None,
        attachment: typing.Optional[str] = None,
        billing_address: typing.Optional[dict] = None,
        due_datetime: typing.Optional[str] = None,
        email_recipients: typing.Optional[list[str]] = None,
        expiration_time: typing.Optional[str] = None,
        extra: typing.Optional[dict] = None,
        generate_qr_code: typing.Optional[bool] = None,
        language: typing.Optional[str] = None,
        mode: typing.Optional[str] = None,
        notifications: typing.Optional[dict] = None,
        order_no: typing.Optional[str] = None,
        product_type: typing.Optional[str] = None,
        redirect_url: typing.Optional[str] = None,
        shopping_address: typing.Optional[dict] = None,
        shortify_attachment_url: typing.Optional[bool] = None,
        shortify_checkout_url: typing.Optional[bool] = None,
        vendor_name: typing.Optional[str] = None,
        webhook_url: typing.Optional[str] = None,
        include_sdk_setup_preload: typing.Optional[bool] = None,
        checkout_extra_args: typing.Optional[dict] = None,
        token: typing.Optional[str] = None,
    ):
        """
        Completes the auto debit flow by automatically
        identifying the "latest" payment method and the token.
        """
        checkout_extra_args = checkout_extra_args or {}
        if not token:
            token = self.get_token_from_db(agreement=agreement, customer_id=customer_id)
        if not pg_codes:
            pg_codes = self.get_auto_debit_pg_codes(
                plugin=txn_type,
                currency=currency_code,
            )
        checkout_response = self.create(
            txn_type=txn_type,
            amount=amount,
            currency_code=currency_code,
            pg_codes=pg_codes,
            payment_type="auto_debit",
            customer_id=customer_id,
            customer_email=customer_email,
            customer_phone=customer_phone,
            customer_first_name=customer_first_name,
            customer_last_name=customer_last_name,
            agreement=agreement,
            card_acceptance_criteria=card_acceptance_criteria,
            attachment=attachment,
            billing_address=billing_address,
            due_datetime=due_datetime,
            email_recipients=email_recipients,
            expiration_time=expiration_time,
            extra=extra,
            generate_qr_code=generate_qr_code,
            language=language,
            mode=mode,
            notifications=notifications,
            order_no=order_no,
            product_type=product_type,
            redirect_url=redirect_url,
            shopping_address=shopping_address,
            shortify_attachment_url=shortify_attachment_url,
            shortify_checkout_url=shortify_checkout_url,
            vendor_name=vendor_name,
            webhook_url=webhook_url,
            include_sdk_setup_preload=include_sdk_setup_preload,
            **checkout_extra_args,
        )
        if not checkout_response["success"]:
            raise APIInterruptError(**checkout_response)
        session_id = checkout_response["response"]["session_id"]
        return self.auto_debit(token=token, session_id=session_id)



from django.conf import settings

# Merchant Related
MERCHANT_ID: str = getattr(settings, "OTTU_MERCHANT_ID")

# Authentication
AUTH: dict = getattr(settings, "OTTU_AUTH")

# Abstract models
ABSTRACT_CHECKOUT_MODEL: bool = getattr(settings, "OTTU_ABSTRACT_CHECKOUT_MODEL", False)
ABSTRACT_WEBHOOK_MODEL: bool = getattr(settings, "OTTU_ABSTRACT_WEBHOOK_MODEL", False)

# Webhook
WEBHOOK_KEY: str = getattr(settings, "OTTU_WEBHOOK_KEY", "")
WEBHOOK_URL: str = getattr(settings, "OTTU_WEBHOOK_URL", "")

# Misc
IS_SANDBOX: bool = getattr(settings, "OTTU_IS_SANDBOX", False)



import typing

import httpx
from httpx import Auth, BasicAuth as _BasicAuth, Request

try:
    from django.core.cache import cache  # pragma: no cover
except ImportError:  # pragma: no cover
    cache = None  # pragma: no cover


class BasicAuth(_BasicAuth):
    def __bool__(self):
        """
        Sample Pseudo Code:
            token = username:password
            encoded_token = encode(token)
            header = f"Basic {encoded_token}"

        Explanation:
            The `token` contains at least 1 character, which is colon `:`.
            The word `Basic` is 5 characters long.
            White space is 1 character long.
            In total, the length of the header is 7 + len(token).
        """
        return len(self._auth_header) > 7


class APIKeyAuth(Auth):
    def __init__(self, api_key: str):
        self.api_key = api_key

    def auth_flow(self, request: Request):
        request.headers["Authorization"] = f"Api-Key {self.api_key}"
        yield request

    def __bool__(self):
        return bool(self.api_key)


class KeycloakAuthBase:
    grant_type: str
    namespace = "keycloak"

    def __init__(
        self,
        host: str,
        realm: str,
        caching: bool = False,
        cache_key="acc_tok",
        *args,
        **kwargs,
    ):
        self.host = host
        self.realm = realm
        self.caching = caching
        self.cache_key = cache_key

    @property
    def token_url(self) -> str:
        return (
            f"https://{self.host}/auth/realms/"
            f"{self.realm}/protocol/openid-connect/token"
        )

    @property
    def cache_key_full(self) -> str:
        return f"{self.namespace}:{self.realm}:{self.cache_key}"

    def get_token_request_payload(self) -> dict:
        return {"grant_type": self.grant_type}

    def _get_token(self):
        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        data = self.get_token_request_payload()
        response = httpx.post(url=self.token_url, data=data, headers=headers)
        response.raise_for_status()
        response_json = response.json()
        access_token = response_json["access_token"]
        ttl = response_json["expires_in"]
        return access_token, ttl

    def get_token_from_cache(self):
        if cache is None:
            # During unittests, the `cache` is always available
            return None  # pragma: no cover
        return cache.get(self.cache_key_full)

    def set_token_in_cache(self, token: str, ttl: int):
        if cache is None:
            # During unittests, the `cache` is always available
            return None  # pragma: no cover
        cache.set(self.cache_key_full, token, timeout=ttl)

    def get_token(self) -> str:
        if self.caching:
            token = self.get_token_from_cache()
            if token:
                return token
        token, ttl = self._get_token()
        if self.caching:
            self.set_token_in_cache(token, ttl)
        return token

    def auth_flow(self, request: Request):
        access_token = self.get_token()
        request.headers["Authorization"] = f"Bearer {access_token}"
        request.headers["X-Service-ID"] = self.realm
        yield request


class KeycloakPasswordAuth(KeycloakAuthBase, Auth):
    grant_type = "password"

    def __init__(
        self,
        username: str,
        password: str,
        client_id: str,
        client_secret: typing.Optional[str] = None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self.username = username
        self.password = password
        self.client_id = client_id
        self.client_secret = client_secret

    def get_token_request_payload(self) -> dict:
        payload = super().get_token_request_payload()
        payload["username"] = self.username
        payload["password"] = self.password
        payload["client_id"] = self.client_id
        if self.client_secret:
            payload["client_secret"] = self.client_secret
        return payload


class KeycloakClientAuth(KeycloakAuthBase, Auth):
    grant_type = "client_credentials"

    def __init__(self, client_id: str, client_secret: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client_id = client_id
        self.client_secret = client_secret

    def get_token_request_payload(self) -> dict:
        payload = super().get_token_request_payload()
        payload["client_id"] = self.client_id
        payload["client_secret"] = self.client_secret
        return payload



import hashlib
import hmac


def remove_empty_values(d: dict):
    """
    Removes empty values from a dictionary.
    """
    return {k: v for k, v in d.items() if v}


def verify_signature(
    payload: dict,
    signature: str,
    webhook_key: str,
) -> bool:
    calculated_signature = calculate_hmac_signature(
        payload=payload,
        hmac_key=webhook_key,
    )
    return calculated_signature == signature


def calculate_hmac_signature(payload: dict, hmac_key: str) -> str:
    # List of fields that are considered for the HMAC signature
    keys = [
        "amount",
        "currency_code",
        "customer_first_name",
        "customer_last_name",
        "customer_email",
        "customer_phone",
        "customer_address_line1",
        "customer_address_line2",
        "customer_address_city",
        "customer_address_state",
        "customer_address_country",
        "customer_address_postal_code",
        "gateway_name",
        "gateway_account",
        "order_no",
        "reference_number",
        "result",
        "state",
    ]

    # Extract and sort the payload keys based on the 'keys' list,
    # and ignore any missing or empty string values
    message = [(k, payload[k]) for k in sorted(payload) if k in keys and payload[k]]

    # Concatenate the key-value pairs
    message_str = "".join([f"{k}{v}" for (k, v) in message])

    # Compute the HMAC signature
    digest = hmac.new(
        bytes(hmac_key, encoding="utf8"),
        bytes(message_str, encoding="utf8"),
        digestmod=hashlib.sha256,
    ).hexdigest()

    return digest



import typing

from . import urls
from .enums import HTTPMethod
from .request import OttuPYResponse
from .utils import remove_empty_values

if typing.TYPE_CHECKING:
    from .ottu import Ottu


class Card:
    def __init__(self, ottu: "Ottu"):
        self.ottu = ottu

    def _get_cards(
        self,
        customer_id: typing.Optional[str] = None,
        pg_codes: typing.Optional[typing.List] = None,
        agreement_id: typing.Optional[str] = None,
    ) -> OttuPYResponse:
        if customer_id is None:
            customer_id = self.ottu.customer_id
        payload = {
            "type": self.ottu.env_type,
            "customer_id": customer_id,
            "pg_codes": pg_codes,
            "agreement_id": agreement_id,
        }
        payload = remove_empty_values(payload)
        return self.ottu.send_request(
            path=urls.USER_CARDS,
            method=HTTPMethod.POST,
            json=payload,
        )

    def get_cards(
        self,
        customer_id: typing.Optional[str] = None,
        pg_codes: typing.Optional[typing.List] = None,
        agreement_id: typing.Optional[str] = None,
    ) -> dict:
        ottu_py_response = self._get_cards(
            customer_id=customer_id,
            pg_codes=pg_codes,
            agreement_id=agreement_id,
        )
        return ottu_py_response.as_dict()

    def list(
        self,
        customer_id: typing.Optional[str] = None,
        pg_codes: typing.Optional[typing.List] = None,
        agreement_id: typing.Optional[str] = None,
    ) -> dict:
        return self.get_cards(
            customer_id=customer_id,
            pg_codes=pg_codes,
            agreement_id=agreement_id,
        )

    def get(
        self,
        customer_id: typing.Optional[str] = None,
        pg_codes: typing.Optional[typing.List] = None,
        agreement_id: typing.Optional[str] = None,
    ) -> typing.Optional[dict]:
        ottu_py_response = self._get_cards(
            customer_id=customer_id,
            pg_codes=pg_codes,
            agreement_id=agreement_id,
        )
        if ottu_py_response.success:
            try:
                return ottu_py_response.response[0]
            except IndexError:
                pass
        return None

    def delete(
        self,
        token: str,
        customer_id: typing.Optional[str] = None,
    ) -> dict:
        customer_id = customer_id or self.ottu.customer_id
        ottu_py_response = self.ottu.send_request(
            path=f"{urls.USER_CARDS}{token}/",
            method=HTTPMethod.DELETE,
            params={
                "customer_id": customer_id,
                "type": self.ottu.env_type,
            },
        )
        return ottu_py_response.as_dict()

    def __repr__(self):
        return f"Card({self.ottu.session.customer_id})"



from dataclasses import asdict


class ResponseMixin:
    def __init__(
        self,
        success: bool,
        status_code: int,
        endpoint: str,
        response: dict,
        error: dict,
    ):
        self.success = success
        self.status_code = status_code
        self.endpoint = endpoint
        self.response = response
        self.error = error

    def as_dict(self) -> dict:
        return {
            "success": self.success,
            "status_code": self.status_code,
            "endpoint": self.endpoint,
            "response": self.response,
            "error": self.error,
        }


class AsDictMixin:
    def as_dict(self) -> dict:
        return asdict(self)  # type: ignore



# Generated by Django 4.2 on 2023-11-17 11:12

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Checkout",
            fields=[
                (
                    "session_id",
                    models.CharField(
                        max_length=100,
                        primary_key=True,
                        serialize=False,
                        verbose_name="Session ID",
                    ),
                ),
                ("type", models.CharField(max_length=250, verbose_name="Type")),
                (
                    "payment_type",
                    models.CharField(max_length=250, verbose_name="Payment Type"),
                ),
                ("amount", models.CharField(max_length=20, verbose_name="Amount")),
                (
                    "currency_code",
                    models.CharField(max_length=10, verbose_name="Currency Code"),
                ),
                ("state", models.CharField(max_length=250, verbose_name="State")),
                (
                    "customer_id",
                    models.CharField(
                        blank=True,
                        max_length=250,
                        verbose_name="Customer ID",
                    ),
                ),
                (
                    "token",
                    models.CharField(blank=True, max_length=250, verbose_name="Token"),
                ),
                (
                    "agreement",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        verbose_name="Agreement",
                    ),
                ),
                (
                    "extra_params",
                    models.JSONField(
                        blank=True,
                        default=dict,
                        verbose_name="Extra Params",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(auto_now_add=True, verbose_name="Created At"),
                ),
                (
                    "updated_at",
                    models.DateTimeField(auto_now=True, verbose_name="Updated At"),
                ),
            ],
            options={
                "verbose_name": "Checkout",
                "verbose_name_plural": "Checkout",
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="Webhook",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "session_id",
                    models.CharField(max_length=250, verbose_name="Session ID"),
                ),
                (
                    "payload",
                    models.JSONField(blank=True, default=dict, verbose_name="Payload"),
                ),
                (
                    "timestamp",
                    models.DateTimeField(auto_now_add=True, verbose_name="Timestamp"),
                ),
                (
                    "checkout",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="ottu.checkout",
                        verbose_name="Checkout",
                    ),
                ),
            ],
            options={
                "verbose_name": "Webhook",
                "verbose_name_plural": "Webhooks",
                "abstract": False,
            },
        ),
    ]



import json
from typing import Any

from .session import PaymentMethod


class PaymentMethodEncoder(json.JSONEncoder):
    def default(self, o: Any) -> Any:
        if isinstance(o, PaymentMethod):
            return o.as_dict()
        return json.JSONEncoder.default(self, o)



from .mixins import ResponseMixin


class OttuBaseError(Exception):
    def __init__(self, msg: str, **extra):
        self.msg = msg

    def __str__(self):
        return self.msg

    def __repr__(self):
        return f"<{self.__class__.__name__}({self.msg})>"


class APIInterruptError(ResponseMixin, OttuBaseError):
    def __str__(self):
        return str(self.error)


class ConfigurationError(OttuBaseError):
    ...


class ValidationError(OttuBaseError):
    ...


class WebhookProcessingError(OttuBaseError):
    ...



import typing

import httpx
from httpx import Auth

from . import urls
from .cards import Card
from .enums import HTTPMethod, TxnType
from .request import OttuPYResponse, RequestResponseHandler
from .session import Session
from .utils import remove_empty_values


class Ottu:
    _session: typing.Optional[Session] = None
    _card: typing.Optional[Card] = None
    session_cls: typing.Type[Session] = Session
    request_response_handler: typing.Type[
        RequestResponseHandler
    ] = RequestResponseHandler

    def __init__(
        self,
        merchant_id: str,
        auth: Auth,
        customer_id: typing.Optional[str] = None,
        is_sandbox: bool = True,
    ) -> None:
        self.merchant_id = merchant_id
        self.host_url = f"https://{merchant_id}"
        self.auth = auth
        self.customer_id = customer_id
        self.is_sandbox = is_sandbox
        self.env_type = "sandbox" if is_sandbox else "production"

        # Other initializations
        self.request_session = self.__create_session()

    def __create_session(self) -> httpx.Client:
        return httpx.Client(auth=self.auth)

    def send_request(
        self,
        path: str,
        method: str,
        **request_params,
    ) -> OttuPYResponse:
        return self.request_response_handler(
            session=self.request_session,
            method=method,
            url=f"{self.host_url}{path}",
            **request_params,
        ).process()

    # Core Methods

    @property
    def session(self):
        if self._session is None:
            self._session = self.session_cls(ottu=self)
        return self._session

    def _update_session(self, session: Session) -> None:
        self._session = session

    def checkout(
        self,
        *,
        txn_type: TxnType,
        amount: str,
        currency_code: str,
        pg_codes: list[str],
        payment_type: str = "one_off",
        customer_id: typing.Optional[str] = None,
        customer_email: typing.Optional[str] = None,
        customer_phone: typing.Optional[str] = None,
        customer_first_name: typing.Optional[str] = None,
        customer_last_name: typing.Optional[str] = None,
        agreement: typing.Optional[dict] = None,
        card_acceptance_criteria: typing.Optional[dict] = None,
        attachment: typing.Optional[str] = None,
        billing_address: typing.Optional[dict] = None,
        due_datetime: typing.Optional[str] = None,
        email_recipients: typing.Optional[list[str]] = None,
        expiration_time: typing.Optional[str] = None,
        extra: typing.Optional[dict] = None,
        generate_qr_code: typing.Optional[bool] = None,
        language: typing.Optional[str] = None,
        mode: typing.Optional[str] = None,
        notifications: typing.Optional[dict] = None,
        order_no: typing.Optional[str] = None,
        product_type: typing.Optional[str] = None,
        redirect_url: typing.Optional[str] = None,
        shopping_address: typing.Optional[dict] = None,
        shortify_attachment_url: typing.Optional[bool] = None,
        shortify_checkout_url: typing.Optional[bool] = None,
        vendor_name: typing.Optional[str] = None,
        webhook_url: typing.Optional[str] = None,
        include_sdk_setup_preload: typing.Optional[bool] = None,
        **kwargs,
    ) -> dict:
        """
        a proxy method to `Session.create(...)`
        """
        return self.session.create(
            txn_type=txn_type,
            amount=amount,
            currency_code=currency_code,
            pg_codes=pg_codes,
            payment_type=payment_type,
            customer_id=customer_id,
            customer_email=customer_email,
            customer_phone=customer_phone,
            customer_first_name=customer_first_name,
            customer_last_name=customer_last_name,
            agreement=agreement,
            card_acceptance_criteria=card_acceptance_criteria,
            attachment=attachment,
            billing_address=billing_address,
            due_datetime=due_datetime,
            email_recipients=email_recipients,
            expiration_time=expiration_time,
            extra=extra,
            generate_qr_code=generate_qr_code,
            language=language,
            mode=mode,
            notifications=notifications,
            order_no=order_no,
            product_type=product_type,
            redirect_url=redirect_url,
            shopping_address=shopping_address,
            shortify_attachment_url=shortify_attachment_url,
            shortify_checkout_url=shortify_checkout_url,
            vendor_name=vendor_name,
            webhook_url=webhook_url,
            include_sdk_setup_preload=include_sdk_setup_preload,
            **kwargs,
        )

    def auto_debit(
        self,
        token: str,
        session_id: str,
    ):
        return self.session.auto_debit(token=token, session_id=session_id)

    @property
    def cards(self) -> Card:
        if self._card is None:
            self._card = Card(ottu=self)
        return self._card

    def checkout_autoflow(
        self,
        *,
        txn_type: TxnType,
        amount: str,
        currency_code: str,
        payment_type: str = "one_off",
        customer_id: typing.Optional[str] = None,
        customer_email: typing.Optional[str] = None,
        customer_phone: typing.Optional[str] = None,
        customer_first_name: typing.Optional[str] = None,
        customer_last_name: typing.Optional[str] = None,
        agreement: typing.Optional[dict] = None,
        card_acceptance_criteria: typing.Optional[dict] = None,
        attachment: typing.Optional[str] = None,
        billing_address: typing.Optional[dict] = None,
        due_datetime: typing.Optional[str] = None,
        email_recipients: typing.Optional[list[str]] = None,
        expiration_time: typing.Optional[str] = None,
        extra: typing.Optional[dict] = None,
        generate_qr_code: typing.Optional[bool] = None,
        language: typing.Optional[str] = None,
        mode: typing.Optional[str] = None,
        notifications: typing.Optional[dict] = None,
        order_no: typing.Optional[str] = None,
        product_type: typing.Optional[str] = None,
        redirect_url: typing.Optional[str] = None,
        shopping_address: typing.Optional[dict] = None,
        shortify_attachment_url: typing.Optional[bool] = None,
        shortify_checkout_url: typing.Optional[bool] = None,
        vendor_name: typing.Optional[str] = None,
        webhook_url: typing.Optional[str] = None,
        include_sdk_setup_preload: typing.Optional[bool] = None,
        checkout_extra_args: typing.Optional[dict] = None,
    ):
        return self.session.checkout_autoflow(
            txn_type=txn_type,
            amount=amount,
            currency_code=currency_code,
            payment_type=payment_type,
            customer_id=customer_id,
            customer_email=customer_email,
            customer_phone=customer_phone,
            customer_first_name=customer_first_name,
            customer_last_name=customer_last_name,
            agreement=agreement,
            card_acceptance_criteria=card_acceptance_criteria,
            attachment=attachment,
            billing_address=billing_address,
            due_datetime=due_datetime,
            email_recipients=email_recipients,
            expiration_time=expiration_time,
            extra=extra,
            generate_qr_code=generate_qr_code,
            language=language,
            mode=mode,
            notifications=notifications,
            order_no=order_no,
            product_type=product_type,
            redirect_url=redirect_url,
            shopping_address=shopping_address,
            shortify_attachment_url=shortify_attachment_url,
            shortify_checkout_url=shortify_checkout_url,
            vendor_name=vendor_name,
            webhook_url=webhook_url,
            include_sdk_setup_preload=include_sdk_setup_preload,
            checkout_extra_args=checkout_extra_args,
        )

    def auto_debit_autoflow(
        self,
        *,
        txn_type: TxnType,
        amount: str,
        currency_code: str,
        customer_id: str,
        agreement: dict,
        pg_codes: typing.Optional[list[str]] = None,
        customer_email: typing.Optional[str] = None,
        customer_phone: typing.Optional[str] = None,
        customer_first_name: typing.Optional[str] = None,
        customer_last_name: typing.Optional[str] = None,
        card_acceptance_criteria: typing.Optional[dict] = None,
        attachment: typing.Optional[str] = None,
        billing_address: typing.Optional[dict] = None,
        due_datetime: typing.Optional[str] = None,
        email_recipients: typing.Optional[list[str]] = None,
        expiration_time: typing.Optional[str] = None,
        extra: typing.Optional[dict] = None,
        generate_qr_code: typing.Optional[bool] = None,
        language: typing.Optional[str] = None,
        mode: typing.Optional[str] = None,
        notifications: typing.Optional[dict] = None,
        order_no: typing.Optional[str] = None,
        product_type: typing.Optional[str] = None,
        redirect_url: typing.Optional[str] = None,
        shopping_address: typing.Optional[dict] = None,
        shortify_attachment_url: typing.Optional[bool] = None,
        shortify_checkout_url: typing.Optional[bool] = None,
        vendor_name: typing.Optional[str] = None,
        webhook_url: typing.Optional[str] = None,
        include_sdk_setup_preload: typing.Optional[bool] = None,
        checkout_extra_args: typing.Optional[dict] = None,
        token: typing.Optional[str] = None,
    ):
        return self.session.auto_debit_autoflow(
            txn_type=txn_type,
            amount=amount,
            currency_code=currency_code,
            customer_id=customer_id,
            pg_codes=pg_codes,
            customer_email=customer_email,
            customer_phone=customer_phone,
            customer_first_name=customer_first_name,
            customer_last_name=customer_last_name,
            agreement=agreement,
            card_acceptance_criteria=card_acceptance_criteria,
            attachment=attachment,
            billing_address=billing_address,
            due_datetime=due_datetime,
            email_recipients=email_recipients,
            expiration_time=expiration_time,
            extra=extra,
            generate_qr_code=generate_qr_code,
            language=language,
            mode=mode,
            notifications=notifications,
            order_no=order_no,
            product_type=product_type,
            redirect_url=redirect_url,
            shopping_address=shopping_address,
            shortify_attachment_url=shortify_attachment_url,
            shortify_checkout_url=shortify_checkout_url,
            vendor_name=vendor_name,
            webhook_url=webhook_url,
            include_sdk_setup_preload=include_sdk_setup_preload,
            checkout_extra_args=checkout_extra_args,
            token=token,
        )

    def raw(
        self,
        method: str,
        path: str,
        headers: typing.Optional[dict] = None,
        **kwargs,
    ):
        """
        To send any sort of http requests to the server.
            method: str - HTTP method name.
                Eg: GET, POST, PUT, DELETE, etc.
            path: str - The path of the request.
                Eg: /b/api/v1/dashboard/statistics
            headers: dict - Optional headers to be sent with the request.
            kwargs: dict - Optional parameters to be sent with the request.
                Supports all the parameters that `httpx.Client.send` supports.
        """
        return self.send_request(
            path=path,
            method=method,
            headers=headers,
            **kwargs,
        )

    def get_payment_methods(
        self,
        plugin,
        currencies: typing.Optional[list[str]] = None,
        customer_id: typing.Optional[str] = None,
        operation: typing.Optional[str] = None,
        tokenizable: bool = False,
        pg_names: typing.Optional[list[str]] = None,
    ) -> dict:
        return self._get_payment_methods(
            plugin=plugin,
            currencies=currencies,
            customer_id=customer_id,
            operation=operation,
            tokenizable=tokenizable,
            pg_names=pg_names,
        ).as_dict()

    def _get_payment_methods(
        self,
        plugin,
        currencies: typing.Optional[list[str]] = None,
        customer_id: typing.Optional[str] = None,
        operation: typing.Optional[str] = None,
        tokenizable: bool = False,
        pg_names: typing.Optional[list[str]] = None,
    ) -> OttuPYResponse:
        payload = {
            "plugin": plugin,
            "currencies": currencies,
            "customer_id": customer_id,
            "operation": operation,
            "tokenizable": tokenizable,
            "pg_names": pg_names,
            "type": "sandbox" if self.is_sandbox else "production",
        }
        payload = remove_empty_values(payload)
        return self.send_request(
            path=urls.PAYMENT_METHODS,
            method=HTTPMethod.POST,
            json=payload,
        )

    def __repr__(self):
        return f"Ottu({self.merchant_id})"



from django.core.exceptions import ImproperlyConfigured
from django.utils.module_loading import import_string

from ....ottu import Ottu as _Ottu
from .. import conf
from ..models import Checkout
from .session import Session


class Ottu(_Ottu):
    model = Checkout
    session_cls = Session

    def get_or_create_session(self):
        instance, _ = self.model.objects.get_or_create(
            session_id=self.session.session_id,
        )
        return instance

    def _create_or_update_dj_session(self):
        session_obj = self.get_or_create_session()
        for field, value in self.session.as_dict().items():
            setattr(session_obj, field, value)
        session_obj.save()

    def _update_session(self, session: Session):  # type: ignore[override]
        super()._update_session(session)
        self._create_or_update_dj_session()


def _generate_instance():
    try:
        auth_conf = conf.AUTH.copy()
        auth_cls = import_string(auth_conf["class"])
    except AttributeError:
        raise ImproperlyConfigured(
            "Must set `OTTU_AUTH` in the settings file",
        )
    except KeyError:
        raise ImproperlyConfigured(
            "The 'class' key is required in the 'AUTH' dictionary",
        )
    except ImportError:
        raise ImproperlyConfigured("The 'class' key is not a valid import path")
    auth_conf.pop("class")
    auth_instance = auth_cls(**auth_conf)
    return Ottu(
        merchant_id=conf.MERCHANT_ID,
        auth=auth_instance,
        is_sandbox=conf.IS_SANDBOX,
    )


ottu = _generate_instance()



from django.db import models
from django.utils.translation import gettext_lazy as _

from . import conf


class Checkout(models.Model):
    session_id = models.CharField(_("Session ID"), max_length=100, primary_key=True)
    type = models.CharField(_("Type"), max_length=250)
    payment_type = models.CharField(_("Payment Type"), max_length=250)
    amount = models.CharField(_("Amount"), max_length=20)
    currency_code = models.CharField(_("Currency Code"), max_length=10)
    state = models.CharField(_("State"), max_length=250)
    customer_id = models.CharField(_("Customer ID"), max_length=250, blank=True)
    token = models.CharField(_("Token"), max_length=250, blank=True)
    agreement = models.JSONField(_("Agreement"), blank=True, default=dict)
    extra_params = models.JSONField(_("Extra Params"), blank=True, default=dict)
    created_at = models.DateTimeField(_("Created At"), auto_now_add=True)
    updated_at = models.DateTimeField(_("Updated At"), auto_now=True)

    class Meta:
        verbose_name = _("Checkout")
        verbose_name_plural = _("Checkout")
        abstract = conf.ABSTRACT_CHECKOUT_MODEL

    def __str__(self):
        return str(self.session_id)


class Webhook(models.Model):
    session_id = models.CharField(_("Session ID"), max_length=250)
    checkout = models.ForeignKey(
        Checkout,
        verbose_name=_(
            "Checkout",
        ),
        on_delete=models.CASCADE,
        blank=True,
        null=True,
    )
    payload = models.JSONField(_("Payload"), blank=True, default=dict)
    timestamp = models.DateTimeField(_("Timestamp"), auto_now_add=True)

    class Meta:
        verbose_name = _("Webhook")
        verbose_name_plural = _("Webhooks")
        abstract = conf.ABSTRACT_WEBHOOK_MODEL

    def __str__(self):
        return str(self.session_id)

    @classmethod
    def create_from_webhook(cls, data):
        session_id = data.get("session_id", "")

        checkout = None
        if session_id:
            checkout = Checkout.objects.filter(session_id=session_id).first()

        instance = cls.objects.create(
            session_id=session_id,
            checkout=checkout,
            payload=data,
        )

        if checkout:
            for field, value in data.items():
                setattr(checkout, field, value)
            checkout.save()
        return instance



import logging
from json import JSONDecodeError
from urllib.parse import urlparse

import httpx

from .mixins import ResponseMixin

logger = logging.getLogger("ottu-py")


class OttuPYResponse(ResponseMixin):
    ...


class RequestResponseHandler:
    def __init__(self, session: httpx.Client, method: str, url: str, **kwargs):
        self.session = session
        self.method = method
        self.url = url
        self.kwargs = kwargs

    @property
    def path(self) -> str:
        """
        The path of the request.
        """
        return urlparse(self.url).path

    def _process_response(self, response: httpx.Response) -> dict:
        if response.status_code == 204:
            return {}

        try:
            result = response.json()
        except JSONDecodeError:
            result = {"detail": response.text}
        return result

    def parse_success_response(
        self,
        httpx_response: httpx.Response,
        parsed_response: dict,
    ) -> OttuPYResponse:
        return OttuPYResponse(
            success=True,
            status_code=httpx_response.status_code,
            endpoint=self.path,
            response=parsed_response,
            error={},
        )

    def parse_non_2xx_error_response(
        self,
        httpx_response: httpx.Response,
        parsed_response: dict,
    ) -> OttuPYResponse:
        return OttuPYResponse(
            success=False,
            status_code=httpx_response.status_code,
            endpoint=self.path,
            response={},
            error=parsed_response,
        )

    def process_response(self, response: httpx.Response) -> OttuPYResponse:
        parsed_response = self._process_response(response)
        if 200 <= response.status_code <= 299:
            return self.parse_success_response(
                httpx_response=response,
                parsed_response=parsed_response,
            )
        return self.parse_non_2xx_error_response(
            httpx_response=response,
            parsed_response=parsed_response,
        )

    def process_unknown_error(self, exc: Exception) -> OttuPYResponse:
        return self._parse_error(exc)

    def process_httpx_error(self, exc: httpx.HTTPError) -> OttuPYResponse:
        return self._parse_error(exc)

    def _parse_error(self, exc: Exception) -> OttuPYResponse:
        return OttuPYResponse(
            success=False,
            status_code=500,
            endpoint=self.path,
            response={},
            error={"detail": str(exc)},
        )

    def _process(self) -> OttuPYResponse:
        try:
            logger.info(
                f"Sending {self.method} request to {self.url} with args {self.kwargs}",
            )
            response = self.session.request(
                method=self.method,
                url=self.url,
                **self.kwargs,
            )
            return self.process_response(response)
        except httpx.HTTPError as exc:
            return self.process_httpx_error(exc)
        except Exception as exc:
            return self.process_unknown_error(exc)

    def process(self) -> OttuPYResponse:
        response = self._process()
        if response.success:
            logger.info(
                f"Received {response.status_code} response "
                f"from {self.url} with args {self.kwargs}",
            )
        else:
            logger.error(
                f"Received {response.status_code} response "
                f"from {self.url} with args {self.kwargs}",
            )
        return response



from django.contrib import admin

from .models import Checkout, Webhook


@admin.register(Checkout)
class CheckoutAdmin(admin.ModelAdmin):
    list_display = (
        "session_id",
        "type",
        "payment_type",
        "amount",
        "currency_code",
        "state",
        "customer_id",
    )
    search_fields = (
        "session_id",
        "type",
        "payment_type",
        "amount",
        "currency_code",
        "state",
        "customer_id",
        "token",
        "agreement",
    )
    list_filter = ["state"]
    ordering = ["-created_at"]


@admin.register(Webhook)
class WebhookAdmin(admin.ModelAdmin):
    list_display = (
        "id",
        "session_id",
        "timestamp",
    )
    search_fields = (
        "session_id",
        "payload",
    )



from .errors import APIInterruptError


def interruption_handler(func):
    """Decorator to handle keyboard interruption."""

    def wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except APIInterruptError as e:
            return e.as_dict()

    return wrapper



from django.apps import AppConfig


class OttuConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "ottu.contrib.django"
    label = "ottu"
    verbose_name = "Ottu"



from ....errors import APIInterruptError
from ....session import Session as _Session
from .. import conf
from ..models import Checkout


class Session(_Session):
    def create(self, *args, **kwargs):
        if conf.WEBHOOK_URL and not kwargs.get("webhook_url"):
            kwargs["webhook_url"] = conf.WEBHOOK_URL
        return super().create(*args, **kwargs)

    def get_token_from_db(self, agreement, customer_id) -> str:
        """
        Get the token from the database
        """
        instance = Checkout.objects.filter(
            customer_id=customer_id,
            agreement__id=agreement.get("id"),
        ).first()
        if instance:
            return instance.token
        raise APIInterruptError(
            success=False,
            status_code=400,
            endpoint="",
            response={},
            error={"detail": "Token not found in the database"},
        )
